from json import JSONDecodeError
from typing import Any

import lxml.etree
import requests
from django import forms
from jmespath.exceptions import LexerError
from lxml.etree import XPathSyntaxError
from lxml.html.soupparser import fromstring

from ..exceptions import CheckError
from . import HttpCheck
from .http import BaseHttpConfig


class XPATHField(forms.CharField):
    def clean(self, value: str) -> str:
        try:
            if value:
                lxml.etree.XPath(value)
        except XPathSyntaxError:
            raise forms.ValidationError("Invalid XPath expression") from None
        return value


class XMLConfig(BaseHttpConfig):
    xpath = XPATHField(required=False)


class XMLCheck(HttpCheck):
    icon = "xml.svg"
    pragma = ["http+xml", "https+xml"]
    config_class = XMLConfig

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        if not cfg.get("url"):
            cfg["url"] = cfg.get("address", "").replace("+xml://", "://")
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            timeout = self.config["timeout"]
            match = self.config["xpath"]
            res = requests.get(self.config["url"], timeout=timeout)
            if res.status_code not in self.config["status_success"]:
                return False
            if match:
                tree = fromstring(res.content)
                if not tree.xpath(self.config["xpath"]):
                    return False
            return True
        except (forms.ValidationError, requests.exceptions.RequestException, JSONDecodeError, LexerError) as e:
            if raise_error:
                raise CheckError("JSON check failed") from e
        return False



from typing import Any

import requests
from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator, URLValidator
from requests.auth import HTTPBasicAuth, HTTPDigestAuth

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm, WriteOnlyField


class SeparatedValuesField(forms.Field):
    def __init__(
        self, base_field: type[forms.Field] = forms.IntegerField, separator: str = ",", *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.base_field = base_field
        self.separator = separator

    def clean(self, data: dict[str, Any]) -> list[int | str]:
        if isinstance(data, str) and self.separator in data:
            self.value_list = data.split(self.separator)
        elif isinstance(data, (list | tuple)):
            self.value_list = data
        else:
            self.value_list = [data]

        base_field = self.base_field()
        return [base_field.clean(value) for value in self.value_list]

    def prepare_value(self, value: list[Any]) -> str:
        if isinstance(value, str):
            return value
        return self.separator.join(map(str, value))


class BaseHttpConfig(ConfigForm):
    url = forms.URLField(assume_scheme="https", validators=[URLValidator()])
    timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(10)], initial=2)
    status_success = SeparatedValuesField(required=True, initial="200")
    username = forms.CharField(required=False)
    password = WriteOnlyField(required=False)
    auth_type = forms.ChoiceField(
        choices=(
            ("", "None"),
            ("basic", "Basic"),
            ("digest", "Digest"),
            ("token", "Token"),
        ),
        required=False,
    )


class HttpConfig(BaseHttpConfig):
    match = forms.CharField(required=False)


class HttpCheck(BaseCheck):
    icon = "http.svg"
    pragma = ["http", "https"]
    config_class = HttpConfig
    address_format: str = "{url}"

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        if not cfg.get("url"):
            cfg["url"] = cfg.get("address", "")
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            timeout = self.config["timeout"]
            match = self.config["match"]
            username, password = self.config["username"], self.config["password"]
            headers = {}
            if self.config["auth_type"] == "basic":
                auth = HTTPBasicAuth(username, password)
            elif self.config["auth_type"] == "digest":
                auth = HTTPDigestAuth(username, password)
            elif self.config["auth_type"] == "token":
                auth = None
                headers = {"Authorization": "access_token myToken"}
            else:
                auth = None
            res = requests.get(self.config["url"], timeout=timeout, auth=auth, headers=headers)
            if res.status_code not in self.config["status_success"]:
                return False
            return not (match and str(match) not in str(res.content))
        except (forms.ValidationError, requests.exceptions.RequestException) as e:
            if raise_error:
                raise CheckError("HTTP check failed") from e
        return False



from django import forms
from django.contrib.auth.forms import AuthenticationForm

from birder.models import Monitor, Project, User


class DateInput(forms.DateInput):
    input_type = "date"


class LoginForm(AuthenticationForm):
    username = forms.EmailField(label="Email", widget=forms.TextInput(attrs={"autofocus": True}))

    class Meta:
        model = User
        fields = ("username", "password")


class MonitorForm(forms.ModelForm):
    class Meta:
        model = Monitor
        fields = "__all__"


class ProjectForm(forms.ModelForm):
    class Meta:
        model = Project
        fields = "__all__"



import logging
from typing import Any

from django.core.exceptions import ValidationError
from django.core.management import BaseCommand
from strategy_field.utils import fqn

from birder.checks import HealthCheck, parser
from birder.exceptions import CheckError
from birder.models import Environment, Monitor, Project

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    requires_migrations_checks = False
    requires_system_checks = ()

    def handle(self, *args: Any, **options: Any) -> None:
        logging.disable(logging.CRITICAL)

        Monitor.objects.all().delete()
        demo, __ = Project.objects.get_or_create(name="Demo", public=True)
        demo.environments.add(*Environment.objects.all())
        demo.save()
        dev, __ = Environment.objects.get_or_create(name="development")
        Monitor.objects.get_or_create(
            project=demo,
            name="Remote",
            environment=dev,
            defaults={"strategy": fqn(HealthCheck)},
        )
        for url in [
            "https://google.com",
            "redis://localhost:26379",
            "postgres://postgres:@localhost/postgres",
            "ftp://alpineftp:alpineftp@localhost:2221",
            "ldap://admin:password@localhost:1389",
            "mysql://root:password@localhost:23306",
            "https+json://dummyjson.com/c/3029-d29f-4014-9fb4",
            "ssh://user:password@localhost:2222",
            "memcache://localhost:21121",
            "rabbitmq://localhost:25672",
            "smtp://admin@example.com:password@localhost:2560",
            "celery://localhost:26379?broker=redis",
            "tcp://localhost:8000",
            "http+xml://google.com",
        ]:
            try:
                checker, config = parser(url)
                frm = checker.config_class(config)
            except ValidationError as e:
                self.stdout.write(self.style.ERROR(f"{url}: {e}"))
            else:
                m, __ = Monitor.objects.get_or_create(
                    project=demo,
                    environment=dev,
                    name=checker.pragma[0],
                    strategy=fqn(checker),
                    defaults={"strategy": fqn(checker), "configuration": config},
                    notes=f"""
## {checker.pragma[0]}

{fqn(checker)}

url: `{url}`


""",
                )
                try:
                    if frm.is_valid():
                        if not checker(configuration=config).check():
                            self.stdout.write(self.style.WARNING(f"{checker.__name__}: {frm.cleaned_data}"))
                        else:
                            self.stdout.write(self.style.SUCCESS(f"{checker.__name__}: {frm.errors}"))
                    else:
                        self.stdout.write(self.style.ERROR(f"{checker.__name__}: {frm.errors}"))

                except (CheckError, ValueError):
                    self.stdout.write(self.style.ERROR(f"{checker.__name__}: {frm.errors}"))



from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator
from pymemcache.client.base import Client as MemCacheClient

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm


class MemCacheConfig(ConfigForm):
    host = forms.CharField(required=True, help_text="Server hostname or IP Address")
    port = forms.IntegerField(validators=[MinValueValidator(1)], initial=11211)
    connect_timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)
    timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)


class MemCacheCheck(BaseCheck):
    icon = "memcache.svg"
    pragma = ["memcache"]
    config_class = MemCacheConfig
    address_format = "{host}:{port}"

    def check(self, raise_error: bool = False) -> bool:
        try:
            base = {**self.config}
            host, port = base.pop("host"), base.pop("port")
            cfg = {"server": (host, port), **base}
            client = MemCacheClient(**cfg)
            client._connect()
            return True
        except ConnectionError as e:
            if raise_error:
                raise CheckError("Redis check failed") from e
        return False



import json
import logging
from datetime import date, datetime
from json import JSONEncoder as JSONEncoder_
from typing import TYPE_CHECKING, Any

import channels.layers
from asgiref.sync import async_to_sync
from constance import config
from django.urls.base import reverse
from strategy_field.utils import fqn

from .consumers import GROUP

if TYPE_CHECKING:
    from birder.models import Monitor

logger = logging.getLogger(__name__)


def notify_ui(msg: str, *args: Any, **kwargs: Any) -> None:
    if msg == "ping":
        _ping(kwargs["timestamp"])
    elif msg == "update":
        _update(*args, **kwargs)
    elif msg == "refresh":
        _refresh(**kwargs)


def _refresh(monitor: "Monitor", crud: str) -> None:
    channel_layer = channels.layers.get_channel_layer()
    async_to_sync(channel_layer.group_send)(
        GROUP,
        {
            "type": "send.json",
            "reason": "update",
            "crud": crud,
        },
    )


def _ping(timestamp: str) -> None:
    channel_layer = channels.layers.get_channel_layer()
    async_to_sync(channel_layer.group_send)(
        GROUP,
        {"type": "send.json", "reason": "ping", "ts": timestamp},
    )


class JSONEncoder(JSONEncoder_):
    def default(self, obj: Any) -> Any:
        from birder.models import Monitor

        if isinstance(obj, Monitor):
            return {
                "id": obj.id,
                "project": {
                    "id": obj.project.id,
                    "data": json.loads(json.dumps(obj.project.data, cls=JSONEncoder)),
                    "status": json.loads(json.dumps(obj.project.status, cls=JSONEncoder)),
                },
                "url": reverse("monitor-detail", args=[obj.pk]),
                "status": obj.status,
                "active": obj.active,
                "name": obj.name,
                "last_check": json.loads(json.dumps(obj.last_timestamp_check, cls=JSONEncoder)),
                "last_error": json.loads(json.dumps(obj.last_timestamp_failure, cls=JSONEncoder)),
                "last_success": json.loads(json.dumps(obj.last_timestamp_success, cls=JSONEncoder)),
                "fqn": fqn(obj.strategy),
                "icon": obj.icon,
                "failures": obj.failures,
                "thresholds": [obj.warn_threshold, obj.err_threshold],
            }
        if isinstance(obj, datetime):
            return obj.strftime(config.DATETIME_FORMAT)
        if isinstance(obj, date):
            return obj.strftime(config.DATE_FORMAT)
        return json.JSONEncoder.default(self, obj)


def _update(monitor: "Monitor") -> None:
    channel_layer = channels.layers.get_channel_layer()
    async_to_sync(channel_layer.group_send)(
        GROUP,
        {"type": "send.json", "reason": "status", "monitor": json.loads(json.dumps(monitor, cls=JSONEncoder))},
    )



from functools import cached_property
from urllib.parse import urlparse

from strategy_field.registry import Registry
from strategy_field.utils import fqn

from .base import BaseCheck


class CheckRegistry(Registry):
    def get_name(self, entry: "BaseCheck") -> str:
        return entry.verbose_name or entry.__name__

    def register(self, check: type[BaseCheck]) -> None:
        super().register(check)

    def as_choices(self) -> list[tuple[str, str]]:
        if not self._choices:
            self._choices = sorted([(fqn(klass), self.get_name(klass)) for klass in self], key=lambda x: x[1])
        return self._choices

    @cached_property
    def protocols(self) -> dict[str, type[BaseCheck]]:
        self._protocols = {}
        for entry in self:
            for p in entry.pragma:
                self._protocols[p.lower()] = entry
        return self._protocols

    def checker_from_url(self, uri: str) -> type[BaseCheck]:
        o = urlparse(uri.strip())
        try:
            checker: type[BaseCheck] = self.protocols[o.scheme]
        except KeyError as e:
            raise ValueError(
                f"{uri} - Unknown protocol '{o.scheme}'. Valid protocols are {list(self.protocols.keys())}"
            ) from e
        return checker


registry = CheckRegistry(BaseCheck)



import logging
from ftplib import FTP
from typing import Any

from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm, WriteOnlyField

logger = logging.getLogger(__name__)


class FtpConfig(ConfigForm):
    host = forms.CharField(required=True)
    port = forms.IntegerField(validators=[MinValueValidator(1)], initial=21)
    timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)
    user = forms.CharField(required=False)
    passwd = WriteOnlyField(required=False)


class FtpCheck(BaseCheck):
    icon = "ftp.svg"
    pragma = ["ftp", "ftps"]
    config_class = FtpConfig
    address_format = "{host}:{port}"

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        if not cfg.get("user"):
            cfg["user"] = cfg.get("username", "")
        if not cfg.get("passwd"):
            cfg["passwd"] = cfg.get("password", "")
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            cfg = {**self.config}
            p = cfg.pop("port")

            ftp = FTP(**cfg)  # noqa: S321
            ftp.port = p
            ftp.connect()
            return True
        except (EOFError, TimeoutError, ConnectionRefusedError) as e:
            if raise_error:
                raise CheckError("FTP check failed") from e
        return False



import logging
import os
from argparse import ArgumentParser
from typing import Any

from constance import config
from django.core.cache import cache
from django.core.management import BaseCommand, call_command
from django.db import IntegrityError

logger = logging.getLogger(__name__)

KEY = "birder:upgrade"


class Command(BaseCommand):
    requires_migrations_checks = False
    requires_system_checks = ()

    def add_arguments(self, parser: ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument(
            "--force",
            action="store_true",
            help="Bypass and delete lock",
        )

    def handle(self, *args: Any, **options: Any) -> None:
        from birder.models import Environment, User

        redis_client = cache.client.get_client()
        if options["force"]:
            redis_client.delete(KEY)
        if redis_client.get(KEY):
            return
        if redis_client.set(KEY, "locked", nx=True, ex=86400):
            try:
                call_command("migrate", interactive=False)
                call_command("collectstatic", interactive=False)
                from django.contrib.auth.models import Group

                Environment.objects.get_or_create(name="development")
                Environment.objects.get_or_create(name="staging")
                Environment.objects.get_or_create(name="production")

                g, is_new = Group.objects.get_or_create(name="Default")
                if is_new:
                    config.NEW_USER_DEFAULT_GROUP = g.pk
                if (admin_user := os.environ.get("ADMIN_USER")) and os.environ.get("ADMIN_PASSWORD"):
                    try:
                        User.objects.create_superuser(
                            username=admin_user, email=admin_user, password=os.environ.get("ADMIN_PASSWORD")
                        )
                        self.stdout.write(self.style.SUCCESS("Superuser created!"))
                    except IntegrityError:
                        self.stdout.write(self.style.WARNING("Exiting superuser found."))

            finally:
                redis_client.delete(KEY)
        else:
            self.stdout.write("Concurrent process detected.")



# Generated by Django 5.1.4 on 2025-01-23 08:22

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("birder", "0002_alter_datahistory_options_alter_environment_options_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="monitor",
            name="verbosity",
        ),
    ]



import logging
from datetime import datetime, timedelta

import dramatiq
from constance import config
from django.utils import timezone
from dramatiq_crontab import cron
from durations_nlp import Duration

from birder.db import DataStore
from birder.models import LogCheck, Monitor
from birder.ws.utils import notify_ui

logger = logging.getLogger(__name__)


@dramatiq.actor
def queue_trigger(pk: str | int) -> None:
    try:
        m = Monitor.objects.get(active=True, pk=pk)
        m.run()
    except Monitor.DoesNotExist:
        logger.warning(f"Monitor #{pk} does not exist")


@cron("*/1 * * * *")  # every 5 minutes
@dramatiq.actor
def process() -> None:
    m: Monitor
    notify_ui("ping", timestamp=timezone.now().strftime(config.DATETIME_FORMAT))
    for m in (
        Monitor.objects.select_related("project", "environment").filter(active=True).order_by("project", "environment")
    ):
        queue_trigger.send(m.pk)


@dramatiq.actor
def clean_log() -> None:
    seconds = Duration(config.RETENTION_POLICY).to_seconds()
    offset = timezone.now() - timedelta(seconds=seconds)
    LogCheck.objects.filter(timestamp__lte=offset).delete()


@dramatiq.actor
def store_history() -> None:
    m: Monitor
    for m in Monitor.objects.all():
        db = DataStore(m)
        db.archive(datetime.now())



from typing import Any

from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator
from kombu import Connection as KombuConnection
from kombu.exceptions import OperationalError

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm, WriteOnlyField


class AmqpConfig(ConfigForm):
    hostname = forms.CharField(required=True, help_text="AMQP Server hostname or IP Address")
    port = forms.IntegerField(validators=[MinValueValidator(1)], initial=5672)
    ssl = forms.BooleanField(required=False, initial=False)
    connect_timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)
    userid = forms.CharField(required=False)
    password = WriteOnlyField(required=False)


class AmqpCheck(BaseCheck):
    icon = "rabbitmq.svg"
    pragma = ["rabbitmq", "amqp", "rabbit"]
    config_class = AmqpConfig
    address_format = "{hostname}:{port}"

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        if not cfg.get("hostname", ""):
            cfg["hostname"] = cfg.get("host", "")
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            conn = KombuConnection(**self.config)
            conn.ensure_connection(max_retries=1)
            return True
        except OperationalError as e:
            if raise_error:
                raise CheckError("AMQP check failed") from e
        return False


RabbitCheck = RabbitMQCheck = KombuCheck = AmqpCheck



from typing import Any

from django.conf import settings
from django.contrib.auth.models import User


def configure_user(user: User | None = None, **kwargs: Any) -> dict[str, Any]:
    if user and user.email in settings.SUPERUSERS:
        user.is_staff = True
        user.is_superuser = True
        user.save()
    return {}



from django.apps import AppConfig


class Config(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "birder"

    def ready(self) -> None:
        from . import handlers  # noqa
        from . import tasks  # noqa



# Generated by Django 5.1.4 on 2025-01-22 19:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("birder", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="datahistory",
            options={"ordering": ["-date"]},
        ),
        migrations.AlterModelOptions(
            name="environment",
            options={"ordering": ["name"]},
        ),
        migrations.AlterModelOptions(
            name="project",
            options={"ordering": ["name"]},
        ),
        migrations.RemoveField(
            model_name="monitor",
            name="env",
        ),
        migrations.RemoveField(
            model_name="project",
            name="default",
        ),
        migrations.AddField(
            model_name="monitor",
            name="environment",
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to="birder.environment"),
        ),
        migrations.AddField(
            model_name="project",
            name="environments",
            field=models.ManyToManyField(related_name="projects", to="birder.environment"),
        ),
        migrations.AddField(
            model_name="project",
            name="icon",
            field=models.CharField(blank=True, default="", max_length=255),
        ),
    ]



"""
ASGI config for config project.

It exposes the ASGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/5.1/howto/deployment/asgi/
"""

import os

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.security.websocket import AllowedHostsOriginValidator
from django.core.asgi import get_asgi_application

from birder.ws.routing import websocket_urlpatterns

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "birder.config.settings")
application = ProtocolTypeRouter(
    {
        "http": get_asgi_application(),
        "websocket": AllowedHostsOriginValidator(AuthMiddlewareStack(URLRouter(websocket_urlpatterns))),
    }
)



import logging
from typing import Any

from django import forms
from django.core.validators import MinValueValidator
from pexpect import pxssh

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm, WriteOnlyField

logger = logging.getLogger(__name__)


class SSHConfig(ConfigForm):
    server = forms.CharField(required=True)
    port = forms.IntegerField(validators=[MinValueValidator(1)], initial=22)
    username = forms.CharField(required=False)
    password = WriteOnlyField(required=False)
    login_timeout = forms.IntegerField(initial=2)


class SSHCheck(BaseCheck):
    icon = "ssh.svg"
    pragma = ["ssh"]
    config_class = SSHConfig
    address_format = "{server}:{port}"

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        cfg["server"] = cfg.get("host", "")
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            s = pxssh.pxssh()
            return s.login(**self.config)
        except pxssh.ExceptionPexpect as e:
            if raise_error:
                raise CheckError("SSH check failed") from e
        return False



from __future__ import annotations

from typing import Any

from environ import Env

ConfigItem = tuple[Any, str | list | bool]

CONFIG: dict[str, ConfigItem] = {
    "ALLOWED_HOSTS": (list, []),
    "AZURE_CLIENT_KEY": (str, ""),
    "AZURE_CLIENT_SECRET": (str, ""),
    "AZURE_TENANT_ID": (str, ""),
    "CACHE_URL": (str, "redis://host:6379/0"),
    "CSRF_TRUSTED_ORIGINS": (list, []),
    "DATABASE_URL": (str, "sqlite:///birder.sqlite3"),
    "DEBUG": (bool, False),
    "EXTRA_APPS": (list, []),
    "GOOGLE_CLIENT_ID": (str, ""),
    "GOOGLE_CLIENT_SECRET": (str, ""),
    "LOG_LEVEL": (str, "ERROR"),
    "SECRET_KEY": (str, "super-secret-key"),
    "SOCIAL_AUTH_LOGIN_URL": (str, "/login/"),
    "SOCIAL_AUTH_RAISE_EXCEPTIONS": (bool, False),
    "SOCIAL_AUTH_REDIRECT_IS_HTTPS": (bool, False),
    "SOCIAL_AUTH_WHITELISTED_DOMAINS": (list, []),
    "STATIC_ROOT": (str, "/app/static/"),
    "STATIC_URL": (str, "static/"),
    "SUPERUSERS": (list, []),
    "VALKEY_URL": (str, "redis://host:6379/0"),
}
env = Env(**CONFIG)
